from enfobench.evaluation.client import ForecastClient
from enfobench.evaluation.evaluate import (
    cross_validate,
    evaluate_metrics,
)

__all__ = [
    "ForecastClient",
    "cross_validate",
    "evaluate_metrics",
]



from datetime import timedelta

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from enfobench.evaluation.filters import hampel
from enfobench.evaluation.utils import periods_in_duration

try:
    import seaborn as sns
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.ticker import LinearLocator, MultipleLocator
    from statsmodels.graphics import tsaplots
except ImportError as e:
    msg = f"Missing optional dependency '{e.name}'. Use pip or conda to install it."
    raise ImportError(msg) from e


def plot_monthly_box(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
) -> tuple[Figure, Axes]:
    """Plot annual seasonality of demand data.

    Args:
        data: Demand data.
        figsize: Figure size.

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()
    data["month"] = data.index.month

    fig, ax = plt.subplots(1, 1, figsize=figsize)

    # Plot the boxplot
    data.boxplot(column="y", by="month", ax=ax)

    # Plot the median values
    data.groupby("month")["y"].median().plot(style="o-", linewidth=0.8, ax=ax)

    # Remove the title and the super title
    plt.title("")
    fig.suptitle("")

    # Set the labels
    ax.set_ylabel("Energy (kWh)")
    ax.set_title("Demand distribution by month", fontsize="large", loc="left")
    return fig, ax


def plot_weekly_box(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
) -> tuple[Figure, Axes]:
    """Plot weekly seasonality of demand data.

    Args:
        data: Demand data.
        figsize: Figure size.

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()
    data["dayofweek"] = data.index.dayofweek + 1

    fig, ax = plt.subplots(1, 1, figsize=figsize)

    # Plot the boxplot
    data.boxplot(column="y", by="dayofweek", ax=ax)

    # Plot the median values
    data.groupby("dayofweek")["y"].median().plot(style="o-", linewidth=0.8, ax=ax)

    # Remove the title and the super title
    plt.title("")
    fig.suptitle("")

    # Set the labels
    ax.set_ylabel("Energy (kWh)")
    ax.set_title("Demand distribution by day of week", loc="left", fontsize="large")
    return fig, ax


def plot_daily_box(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
) -> tuple[Figure, Axes]:
    """Plot daily seasonality of demand data.

    Args:
        data: Demand data.
        figsize: Figure size.

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()
    data["hour"] = data.index.hour + 1

    fig, ax = plt.subplots(1, 1, figsize=figsize)

    # Plot the boxplot
    data.boxplot(column="y", by="hour", ax=ax)

    # Plot the median values
    data.groupby("hour")["y"].median().plot(style="o-", linewidth=0.8, ax=ax)

    # Remove the title and the super title
    plt.title("")
    fig.suptitle("")

    # Set the labels
    ax.set_ylabel("Energy (kWh)")
    ax.set_title("Demand distribution by hour", loc="left", fontsize="large")
    return fig, ax


def plot_histogram(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
    n_bins: int = 100,
) -> tuple[Figure, Axes]:
    """Plot a histogram of demand data.

    Args:
            data: Demand data.
            figsize: Figure size.
            n_bins: Number of bins for the histogram.

        Returns:
            fig: Figure object.
            ax: Axes object.

    """
    # define the energy intervals to use for the histogram
    bins = np.linspace(0, data.y.max(), n_bins + 1)
    # calculate the distribution of energy values
    hist_df, _ = np.histogram(data["y"], bins=bins)

    fig, ax = plt.subplots(1, 2, figsize=figsize)

    # plot the histogram
    ax[0].plot(bins[:-1], hist_df)
    ax[0].set_xlabel("Energy (kWh)")
    ax[0].set_ylabel("Count")
    ax[0].grid()

    # Set title
    ax[0].set_title("Demand distribution", loc="left", fontsize="large")

    # plot the histogram on a logarithmic scale
    ax[1].plot(bins[:-1], hist_df)
    ax[1].set_yscale("log")
    ax[1].set_ylim(10, None)
    ax[1].set_xlabel("Energy (kWh)")
    ax[1].set_ylabel("Count")
    ax[1].grid(which="both")
    return fig, ax


def plot_heatmap(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
    **kwargs,
) -> tuple[Figure, Axes]:
    """Plot a heatmap of demand data.

    Args:
        data: Demand data.
        figsize: Figure size.
        **kwargs: Additional arguments to pass to the seaborn heatmap function.

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()

    # Add columns for date and time
    data["date"] = data.index.date
    data["time"] = data.index.time

    # Pivot dates and times to create a two-dimensional representation
    data_hm = data.pivot_table(index="time", columns="date", values="y", aggfunc=lambda x: x.iloc[0], dropna=False)

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.set_title("Demand heatmap", loc="left", fontsize="large")

    sns.heatmap(data_hm, ax=ax, cbar_kws={"label": "Energy (kWh)"}, **kwargs)
    return fig, ax


def plot_3d(data: pd.DataFrame, figsize: tuple[float, float] = (10, 8)) -> tuple[Figure, Axes]:
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()

    # Add columns for date and time
    data["date"] = data.index.date
    data["time"] = data.index.time

    # Pivot dates and times to create a two-dimensional representation
    data_hm = data.pivot_table(index="time", columns="date", values="y", aggfunc=lambda x: x.iloc[0], dropna=False)

    fig = plt.figure(figsize=figsize)
    ax = plt.subplot(111, projection="3d")

    # Create meshgrid
    X, Y = np.meshgrid(np.arange(data_hm.shape[1]), np.arange(data_hm.shape[0]))  # noqa: N806

    # Plot the surface
    plot = ax.plot_surface(X=X, Y=Y, Z=np.nan_to_num(data_hm.values), cmap="YlGnBu_r", vmin=0)

    # set tick every three hours for the time axis
    ax.set_yticks(ticks=np.linspace(0, data_hm.shape[0], 8 + 1))
    ax.set_yticklabels(labels=range(0, 25, 3))
    ax.set_ylabel("time")

    # set ticks for the date axis based on the heatmap above
    tick_index = range(0, data_hm.shape[1], int(data_hm.shape[1] / 10))
    tick_labels = [data_hm.columns[i] for i in tick_index]
    ax.set_xticks(ticks=tick_index)
    ax.set_xticklabels(labels=tick_labels, rotation=45)
    ax.set_xlabel("")

    # set z-axis limits and ticks
    ax.set_zlim(0, round(data["y"].max()))
    ax.zaxis.set_major_locator(LinearLocator(11))
    ax.set_zlabel("Energy (kWh)")

    # Remove gray panes and axis grid
    ax.xaxis.pane.fill = False
    ax.xaxis.pane.set_edgecolor("white")
    ax.yaxis.pane.fill = False
    ax.yaxis.pane.set_edgecolor("white")
    ax.zaxis.pane.fill = False
    ax.zaxis.pane.set_edgecolor("white")
    ax.grid(False)

    fig.colorbar(plot, ax=ax, label="Energy (kWh)")
    return fig, ax


def plot_data_quality(data: pd.DataFrame, figsize: tuple[float, float] = (12, 5)) -> tuple[Figure, Axes]:
    """Plot the data quality of demand data.

    Args:
        data: Demand data.
        figsize: Figure size.

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    # Make a copy of the data to avoid modifying the original data
    data = data.copy()
    data["quality"] = 3
    data.loc[data["y"].isna(), "quality"] = 0
    data.loc[data["y"].eq(0), "quality"] = 1
    _, is_outlier = hampel(data["y"])
    data.loc[is_outlier, "quality"] = 2

    # Add columns for date and time
    data["date"] = data.index.date
    data["time"] = data.index.time

    # Pivot dates and times to create a two-dimensional representation
    data_hm = data.pivot_table(
        index="time", columns="date", values="quality", aggfunc=lambda x: x.iloc[0], dropna=False
    )
    data_hm.columns = [str(col) for col in data_hm.columns]

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.set_title("Data quality of demand data", loc="left", fontsize="large")

    cmap = plt.get_cmap("Reds", 4)
    sns.heatmap(
        data_hm,
        ax=ax,
        cmap=cmap,
        cbar_kws={"label": "Quality metric", "ticks": [0.5, 1.5, 2.5, 3.5]},
        vmin=0,
        vmax=4,
    )
    ax.collections[0].colorbar.set_ticklabels(["Missing", "Zero", "Outlier", "Good Data"])
    return fig, ax


def plot_acf(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
    lags: pd.Timedelta | timedelta | str = "7D",
) -> tuple[Figure, Axes]:
    """Plots the autocorrelation function.

    Args:
        data: Demand data.
        figsize: Figure size.
        lags: Number of lags to plot. If a string is passed, it must be a valid
            pandas frequency string (e.g. '7D' for 7 days).

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    periods = periods_in_duration(data.index, duration=lags)
    tsaplots.plot_acf(data["y"], ax=ax, lags=periods)

    plt.title("")
    ax.set_title("Autocorrelation", loc="left", fontsize="large")
    ax.grid()
    ax.set_xlim(0 - periods * 0.01, periods * 1.01)

    periods_in_one_day = periods_in_duration(data.index, duration="1D")
    if periods <= periods_in_one_day:
        periods_in_three_hours = periods_in_duration(data.index, duration="3H")
        ax.xaxis.set_major_locator(MultipleLocator(periods_in_three_hours))
        ax.xaxis.set_major_formatter(lambda x, pos: str(int(x / periods_in_three_hours * 3)))  # noqa
        ax.xaxis.set_minor_locator(MultipleLocator(periods_in_duration(data.index, duration="1H")))
        ax.set_xlabel("Lag (hours)")
    else:
        ax.xaxis.set_major_locator(MultipleLocator(periods_in_one_day))
        ax.xaxis.set_major_formatter(lambda x, pos: str(int(x / periods_in_one_day)))  # noqa
        ax.xaxis.set_minor_locator(MultipleLocator(periods_in_duration(data.index, duration="6H")))
        ax.set_xlabel("Lag (days)")
    return fig, ax


def plot_pacf(
    data: pd.DataFrame,
    figsize: tuple[float, float] = (12, 5),
    lags: pd.Timedelta | timedelta | str = "7D",
) -> tuple[Figure, Axes]:
    """Plots the partial autocorrelation function.

    Args:
        data: Demand data.
        figsize: Figure size.
        lags: Number of lags to plot. If a string is passed, it must be a valid
            pandas frequency string (e.g. '7D' for 7 days).

    Returns:
        fig: Figure object.
        ax: Axes object.
    """
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    periods = periods_in_duration(data.index, duration=lags)
    tsaplots.plot_pacf(data["y"], ax=ax, lags=periods)

    plt.title("")
    ax.set_title("Partial autocorrelation", loc="left", fontsize="large")
    ax.grid()
    ax.set_xlim(0 - periods * 0.01, periods * 1.01)

    periods_in_one_day = periods_in_duration(data.index, duration="1D")
    if periods <= periods_in_one_day:
        periods_in_three_hours = periods_in_duration(data.index, duration="3H")
        ax.xaxis.set_major_locator(MultipleLocator(periods_in_three_hours))
        ax.xaxis.set_major_formatter(lambda x, pos: str(int(x / periods_in_three_hours * 3)))  # noqa
        ax.xaxis.set_minor_locator(MultipleLocator(periods_in_duration(data.index, duration="1H")))
        ax.set_xlabel("Lag (hours)")
    else:
        ax.xaxis.set_major_locator(MultipleLocator(periods_in_one_day))
        ax.xaxis.set_major_formatter(lambda x, pos: str(int(x / periods_in_one_day)))  # noqa
        ax.xaxis.set_minor_locator(MultipleLocator(periods_in_duration(data.index, duration="6H")))
        ax.set_xlabel("Lag (days)")
    return fig, ax



from random import randrange

import numpy as np
import pandas as pd
import pytest

from enfobench import AuthorInfo, Dataset, ForecasterType, ModelInfo
from enfobench.evaluation.utils import create_forecast_index


class TestModel:
    def __init__(self, param1: int):
        self.param1 = param1

    def info(self) -> ModelInfo:
        return ModelInfo(
            name="TestModel",
            authors=[
                AuthorInfo("Author 1", "author-1@institution.org"),
            ],
            type=ForecasterType.point,
            params={
                "param1": 1,
            },
        )

    def forecast(
        self,
        horizon: int,
        history,
        past_covariates=None,  # noqa: ARG002
        future_covariates=None,  # noqa: ARG002
        level=None,  # noqa: ARG002
        metadata=None,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ):
        index = create_forecast_index(history, horizon)
        y_hat = np.full(horizon, fill_value=history["y"].mean()) + self.param1
        return pd.DataFrame(
            index=index,
            data={
                "yhat": y_hat,
            },
        )


@pytest.fixture(scope="function")
def model():
    return TestModel(1)


class Helpers:
    @staticmethod
    def generate_target(**kwargs) -> pd.DataFrame:
        index = pd.date_range(**kwargs)
        target = pd.DataFrame(
            index=index,
            data={
                "y": np.random.random(len(index)),
            },
        )
        return target

    @staticmethod
    def generate_covariates(columns: list[str], **kwargs):
        index = pd.date_range(**kwargs)
        df = pd.DataFrame(
            index=index,
            data={c: np.random.random(len(index)) for c in columns},
        )
        return df

    @staticmethod
    def generate_future_covariates(columns: list[str], **kwargs):
        cutoff_dates = pd.date_range(**kwargs)
        forecasts = []

        for cutoff_date in cutoff_dates:
            index = pd.date_range(cutoff_date + pd.Timedelta(hours=1), cutoff_date + pd.Timedelta(days=7), freq="H")
            forecast = pd.DataFrame(
                data={"timestamp": index, **{c: np.random.random(len(index)) for c in columns}},
            )
            forecast["cutoff_date"] = cutoff_date
            forecasts.append(forecast)

        return pd.concat(forecasts, ignore_index=True)

    @staticmethod
    def generate_univariate_dataset(**kwargs):
        target = Helpers.generate_target(**kwargs)
        return Dataset(target)

    @staticmethod
    def generate_multivariate_dataset(columns: list[str], **kwargs):
        target = Helpers.generate_target(**kwargs)
        past_covariates = Helpers.generate_covariates(columns, **kwargs)
        return Dataset(
            target=target,
            past_covariates=past_covariates,
        )

    @staticmethod
    def random_date(start: pd.Timestamp, end: pd.Timestamp, resolution: str) -> pd.Timestamp:
        """
        This function will return a random datetime between two datetime
        objects.
        """
        resolution_in_seconds = int(pd.Timedelta(resolution).total_seconds())
        delta = end - start
        int_delta = int(delta.total_seconds())
        random_second = randrange(0, int_delta, resolution_in_seconds)  # noqa: S311
        return start + pd.Timedelta(seconds=random_second)


@pytest.fixture(scope="session")
def helpers():
    return Helpers


@pytest.fixture(scope="session")
def clean_forecasts() -> pd.DataFrame:
    cutoff_dates = pd.date_range("2020-01-01", "2021-01-01", freq="D")

    forecasts = []
    for cutoff_date in cutoff_dates:
        index = pd.date_range(cutoff_date, cutoff_date + pd.Timedelta(hours=25), freq="H")
        forecast = pd.DataFrame(
            data={
                "timestamp": index,
                "yhat": np.random.random(len(index)),
                "y": np.random.random(len(index)),
            },
        )
        forecast["cutoff_date"] = cutoff_date
        forecasts.append(forecast)

    forecast_df = pd.concat(forecasts, ignore_index=True)
    assert not forecast_df.isna().any(axis=None)
    return forecast_df


@pytest.fixture(scope="session")
def forecasts_with_missing_values() -> pd.DataFrame:
    cutoff_dates = pd.date_range("2020-01-01", "2021-01-01", freq="D")

    forecasts = []
    for cutoff_date in cutoff_dates:
        index = pd.date_range(cutoff_date, cutoff_date + pd.Timedelta(hours=25), freq="H")
        forecast = pd.DataFrame(
            data={
                "timestamp": index,
                "yhat": np.random.random(len(index)),
                "y": np.random.random(len(index)),
            },
        )
        forecast["cutoff_date"] = cutoff_date
        forecasts.append(forecast)

    forecast_df = pd.concat(forecasts, ignore_index=True)
    forecast_df.loc[forecast_df["y"] <= 0.3, "y"] = np.nan
    assert forecast_df.isna().any(axis=None)
    return forecast_df



import numpy as np
from pandas import Series


def hampel(series: Series, window_size: int = 5, n_sigmas: float = 3.0) -> tuple[Series, Series]:
    """Apply the Hampel filter to a series.

    The Hampel filter is generally used to detect anomalies in data with a timeseries structure.
    It basically consists of a sliding window of a parameterizable size. For each window, each observation
    will be compared with the Median Absolute Deviation (MAD). The observation will be considered an outlier
    in the case in which it exceeds the MAD by n times (the parameter n is also parameterizable).

    Args:
        series: Series to filter.
        window_size: The size of the moving window for outlier detection (default is 5).
        n_sigmas: The number of standard deviations for outlier detection (default is 3.0).
    """
    # make a copy of the series to avoid modifying the original
    series = series.copy()

    scaling_constant = 1.4826  # for Gaussian distribution

    rolling_median = series.rolling(window=window_size, center=True).median()
    rolling_median_absolute_deviation = scaling_constant * series.rolling(window=window_size, center=True).apply(
        lambda x: np.median(np.abs(x - np.median(x)))
    )

    deviation_from_median_absolute_deviation = np.abs(series - rolling_median)

    # select values that deviate more than n_sigma from the mad and replace them with the median
    mask = deviation_from_median_absolute_deviation > (n_sigmas * rolling_median_absolute_deviation)
    series[mask] = rolling_median[mask]
    return series, mask



import logging
import warnings
from abc import ABCMeta
from pathlib import Path

import pandas as pd
from pandas.tseries.frequencies import to_offset

logger = logging.getLogger(__name__)


class Subset(metaclass=ABCMeta):  # noqa: B024
    """Subset class representing one of the subset of the HuggingFace dataset.

    Args:
        file_path: The path to the subset file.
    """

    def __init__(self, file_path: Path | str) -> None:
        file_path = Path(file_path).resolve()
        if not file_path.is_file() or not file_path.exists():
            msg = "Please provide the existing file where the subset is located."
            raise ValueError(msg)
        self.file_path = file_path

    def __repr__(self):
        return f"{self.__class__.__name__}(file_path={self.file_path})"

    def read(self) -> pd.DataFrame:
        """Read the subset from the file."""
        return pd.read_parquet(self.file_path)


class Dataset:
    """Dataset class representing a collection of data required for forecasting task.

    Args:
        target: The target variable.
        past_covariates: The past covariates.
        future_covariates: The future covariates.
        metadata: The metadata.
    """

    def __init__(
        self,
        target: pd.DataFrame,
        past_covariates: pd.DataFrame | None = None,
        future_covariates: pd.DataFrame | None = None,
        metadata: dict | None = None,
    ):
        self._target = self._check_target(target.copy())
        self._first_available_target_date: pd.Timestamp = self._target.index[0]
        self._last_available_target_date: pd.Timestamp = self._target.index[-1]

        self._past_covariates = (
            self._check_past_covariates(past_covariates.copy()) if past_covariates is not None else None
        )

        self._future_covariates = (
            self._check_external_forecasts(future_covariates.copy()) if future_covariates is not None else None
        )
        self.metadata = metadata

    @property
    def target_available_since(self) -> pd.Timestamp:
        """Returns the first available target date."""
        return self._first_available_target_date

    @property
    def target_available_until(self) -> pd.Timestamp:
        """Returns the last available target date."""
        return self._last_available_target_date

    @property
    def target_freq(self) -> str:
        """Returns the frequency of the target."""
        return self._target.index.inferred_freq

    @staticmethod
    def _check_target(y: pd.DataFrame) -> pd.DataFrame:
        if isinstance(y, pd.Series):
            msg = "Target is a Series, converting to DataFrame."
            warnings.warn(msg, UserWarning, stacklevel=2)
            y = y.to_frame("y")

        if not isinstance(y.index, pd.DatetimeIndex):
            msg = f"Target dataframe must have DatetimeIndex, have {y.index.__class__.__name__}."
            raise ValueError(msg)

        if y.index.inferred_freq is None:
            msg = "Frequency cannot be inferred from target's index, resampling target."
            warnings.warn(msg, UserWarning, stacklevel=2)
            freq = to_offset(y.index[1] - y.index[0]).freqstr
            freq = "1" + freq if not freq[0].isdigit() else freq
            y = y.resample(freq).asfreq()

        y.rename_axis("timestamp", inplace=True)
        y.sort_index(inplace=True)
        return y

    def _check_past_covariates(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        if not isinstance(X.index, pd.DatetimeIndex):
            msg = f"Past covariates must have DatetimeIndex, have {X.index.__class__.__name__}."
            raise ValueError(msg)

        X.rename_axis("timestamp", inplace=True)
        X.sort_index(inplace=True)

        if X.index[0] > self._first_available_target_date:
            msg = "Covariates should be provided for the full target timeframe, covariates start after target values."
            warnings.warn(msg, UserWarning, stacklevel=2)

        if X.index[-1] < self._last_available_target_date:
            msg = "Covariates should be provided for the full target timeframe, covariates end before target values."
            warnings.warn(msg, UserWarning, stacklevel=2)
        return X

    def _check_external_forecasts(self, X: pd.DataFrame) -> pd.DataFrame:  # noqa: N803
        first_forecast_date = X.cutoff_date.min()
        last_forecast_date = X.cutoff_date.max()
        last_forecast_end_date = X[X.cutoff_date == last_forecast_date].timestamp.max()

        if first_forecast_date > self._first_available_target_date:
            msg = (
                "External forecasts should be provided for the full target timeframe, "
                "forecasts start after target values."
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        if last_forecast_end_date < self._last_available_target_date:
            msg = (
                "External forecasts should be provided for the full target timeframe, "
                "forecasts end before target values."
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        return X

    def _check_cutoff_in_rage(self, cutoff_date: pd.Timestamp):
        if cutoff_date < self._first_available_target_date:
            msg = f"Cutoff date is before the start date: {cutoff_date} < {self._first_available_target_date}."
            raise IndexError(msg)

        if cutoff_date > self._last_available_target_date:
            msg = f"Cutoff date is after the end date: {cutoff_date} > {self._last_available_target_date}."
            raise IndexError(msg)

    def get_history(self, cutoff_date: pd.Timestamp) -> pd.DataFrame:
        """Returns the history of the target variable up to the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The history of the target variable up to the cutoff date.
        """
        self._check_cutoff_in_rage(cutoff_date)
        return self._target[self._target.index < cutoff_date]

    def get_past_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the past covariates for the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The past covariates up until the cutoff date.
        """
        if self._past_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        return self._past_covariates[self._past_covariates.index < cutoff_date]

    def get_future_covariates(self, cutoff_date: pd.Timestamp) -> pd.DataFrame | None:
        """Returns the future covariates for the cutoff date.

        Args:
            cutoff_date: The cutoff date.

        Returns:
            The last external forecasts made before the cutoff date.
        """
        if self._future_covariates is None:
            return None

        self._check_cutoff_in_rage(cutoff_date)
        last_past_cutoff_date = self._future_covariates.cutoff_date[
            self._future_covariates.cutoff_date <= cutoff_date
        ].max()

        future_covariates = self._future_covariates[
            (self._future_covariates.cutoff_date == last_past_cutoff_date)
            & (self._future_covariates.timestamp >= cutoff_date)
        ]
        future_covariates.set_index("timestamp", inplace=True)
        return future_covariates



from enfobench.datasets.electricity_demand import DemandSubset, ElectricityDemandDataset, MetadataSubset, WeatherSubset

__all__ = [
    "ElectricityDemandDataset",
    "DemandSubset",
    "MetadataSubset",
    "WeatherSubset",
]



import numpy as np
import pandas as pd

from enfobench.datasets.utils import create_perfect_forecasts_from_covariates


def test_create_perfect_forecasts_from_covariates():
    index = pd.date_range(start="2020-01-01", end="2020-10-02 13:54:00", freq="1H")
    past_covariates = pd.DataFrame(
        index=index,
        data=np.random.rand(len(index), 2),
        columns=["covariate_1", "covariate_2"],
    )
    past_covariates.drop(index[2], inplace=True)

    future_covariates = create_perfect_forecasts_from_covariates(
        past_covariates,
        start=pd.Timestamp("2020-01-01"),
        step=pd.Timedelta("1D"),
        horizon=pd.Timedelta("7D"),
    )

    assert isinstance(future_covariates, pd.DataFrame)
    assert "covariate_1" in future_covariates.columns
    assert "covariate_2" in future_covariates.columns
    assert "timestamp" in future_covariates.columns
    assert "cutoff_date" in future_covariates.columns


# def test_create_perfect_forecasts_from_covariates(covariates):
#     forecasts = enfobench.dataset.utils.create_perfect_forecasts_from_covariates(
#         covariates,
#         horizon=pd.Timedelta("7 days"),
#         step=pd.Timedelta("1D"),
#     )
#
#     assert isinstance(forecasts, pd.DataFrame)
#     assert "timestamp" in forecasts.columns
#     assert "cutoff_date" in forecasts.columns
#     assert all(col in forecasts.columns for col in covariates.columns)



from pathlib import Path
from typing import Any

import duckdb
import pandas as pd

from enfobench.core import Subset

Metadata = dict[str, Any]


class MetadataSubset(Subset):
    """Metadata subset of the HuggingFace dataset containing all metadata about the meters.

    Args:
        file_path: The path to the subset file.
    """

    def list_unique_ids(self) -> list[str]:
        """Lists all unique ids."""
        query = """
            SELECT DISTINCT unique_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().unique_id.tolist()

    def get_by_unique_id(self, unique_id: str) -> Metadata:
        """Returns the metadata for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)
        return df.iloc[0].to_dict()


class WeatherSubset(Subset):
    """Weather subset of the HuggingFace dataset containing all weather data.

    Args:
        file_path: The path to the subset file.
    """

    def list_location_ids(self) -> list[str]:
        """Lists all location ids."""
        query = """
            SELECT DISTINCT location_id
            FROM read_parquet(?)
        """
        conn = duckdb.connect(":memory:")
        return conn.execute(query, parameters=[str(self.file_path)]).fetch_df().location_id.tolist()

    def get_by_location_id(self, location_id: str, columns: list[str] | None = None) -> pd.DataFrame:
        """Returns the weather data for the given location id.

        Args:
            location_id: The location id of the weather station.
            columns: The columns to return. If None, all columns are returned.
        """
        conn = duckdb.connect(":memory:")

        if columns:
            query = f"""
                SELECT timestamp, {", ".join(columns)}
                FROM read_parquet(?)
                WHERE location_id = ?
            """  # noqa: S608
        else:
            query = """
                SELECT *
                FROM read_parquet(?)
                WHERE location_id = ?
            """
        df = conn.execute(query, parameters=[str(self.file_path), location_id]).fetch_df()
        if df.empty:
            msg = f"Location id '{location_id}' was not found."
            raise KeyError(msg)

        # Remove location_id and set timestamp as index
        df.drop(columns=["location_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class DemandSubset(Subset):
    """Data subset of the HuggingFace dataset containing all electricity load data.

    Args:
        file_path: The path to the subset file.
    """

    def get_by_unique_id(self, unique_id: str):
        """Returns the demand data for the given unique id.

        Args:
            unique_id: The unique id of the meter.
        """
        query = """
            SELECT *
            FROM read_parquet(?)
            WHERE unique_id = ?
        """
        conn = duckdb.connect(":memory:")
        df = conn.execute(query, parameters=[str(self.file_path), unique_id]).fetch_df()
        if df.empty:
            msg = f"Unique id '{unique_id}' was not found."
            raise KeyError(msg)

        # Remove unique_id and set timestamp as index
        df.drop(columns=["unique_id"], inplace=True, errors="ignore")
        df.set_index("timestamp", inplace=True)
        return df


class ElectricityDemandDataset:
    """ElectricityDemandDataset class representing the HuggingFace dataset.

    This class is a collection of all subsets inside HuggingFace dataset.
    It provides an easy way to access the different subsets.

    Args:
        directory: The directory where the HuggingFace dataset is located.
                   This directory should contain all the subset files.
    """

    HUGGINGFACE_DATASET = "EDS-lab/electricity-demand"
    SUBSETS = ("demand", "metadata", "weather")

    def __init__(self, directory: Path | str) -> None:
        directory = Path(directory).resolve()
        if not directory.is_dir() or not directory.exists():
            msg = f"Please provide the existing directory where the '{self.HUGGINGFACE_DATASET}' dataset is located."
            raise ValueError(msg)
        self.directory = directory.resolve()

    def __repr__(self) -> str:
        return f"DemandDataset(directory={self.directory})"

    def _check_for_valid_subset(self, subset: str):
        if subset not in self.SUBSETS:
            msg = f"Please provide a valid subset. Available subsets: {self.SUBSETS}"
            raise ValueError(msg)

    @property
    def metadata_subset(self) -> MetadataSubset:
        """Returns the metadata subset."""
        return MetadataSubset(self._get_subset_path("metadata"))

    @property
    def weather_subset(self) -> WeatherSubset:
        """Returns the weather subset."""
        return WeatherSubset(self._get_subset_path("weather"))

    @property
    def demand_subset(self) -> DemandSubset:
        """Returns the demand subset."""
        return DemandSubset(self._get_subset_path("demand"))

    def _get_subset_path(self, subset: str) -> Path:
        filepath = self.directory / f"{subset}.parquet"
        if not filepath.exists():
            msg = (
                f"There is no {subset} in the directory. "
                f"Make sure to download all subsets from the HuggingFace dataset: {self.HUGGINGFACE_DATASET}."
            )
            raise ValueError(msg)
        return self.directory / f"{subset}.parquet"

    def list_unique_ids(self) -> list[str]:
        return self.metadata_subset.list_unique_ids()

    def list_location_ids(self) -> list[str]:
        return self.weather_subset.list_location_ids()

    def get_data_by_unique_id(self, unique_id: str) -> tuple[pd.DataFrame, pd.DataFrame, Metadata]:
        metadata = self.metadata_subset.get_by_unique_id(unique_id)
        location_id = metadata["location_id"]

        demand = self.demand_subset.get_by_unique_id(unique_id)
        weather = self.weather_subset.get_by_location_id(location_id)
        return demand, weather, metadata



from enfobench.core.dataset import Dataset, Subset
from enfobench.core.model import AuthorInfo, ForecasterType, Model, ModelInfo

__all__ = [
    "Dataset",
    "Subset",
    "Model",
    "ModelInfo",
    "AuthorInfo",
    "ForecasterType",
]



import numpy as np
from numpy import ndarray


def check_not_empty(*arrays: ndarray) -> None:
    """Check that none of the arrays are not empty.

    Args:
        *arrays: Objects that will be checked for emptiness.
    """
    if any(array.size == 0 for array in arrays):
        msg = "Found empty array in inputs."
        raise ValueError(msg)


def check_consistent_length(*arrays: ndarray) -> None:
    """Check that all arrays have consistent length.

    Checks whether all input arrays have the same length.

    Args:
        *arrays: Objects that will be checked for consistent length.
    """
    if any(array.ndim != 1 for array in arrays):
        mag = "Found multi dimensional array in inputs."
        raise ValueError(mag)

    lengths = [len(array) for array in arrays]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        msg = f"Found input variables with inconsistent numbers of samples: {lengths}"
        raise ValueError(msg)


def check_has_no_nan(*arrays: ndarray) -> None:
    """Check that all arrays have no NaNs.

    Args:
        *arrays: Objects that will be checked for NaNs.
    """
    for array in arrays:
        if np.isnan(array).any():
            msg = f"Found NaNs in input variables: {array.__repr__()}"
            raise ValueError(msg)


def check_arrays(*arrays: ndarray) -> None:
    """Check that all arrays are valid.

    Args:
        *arrays: Objects that will be checked for validity.
    """
    check_not_empty(*arrays)
    check_consistent_length(*arrays)
    check_has_no_nan(*arrays)


def mean_absolute_error(y_true: ndarray, y_pred: ndarray) -> float:
    """Mean absolute error regression loss.

    Args:
        y_true: Ground truth (correct) target values.
        y_pred: Estimated target values.
    """
    check_arrays(y_true, y_pred)
    return float(np.mean(np.abs(y_true - y_pred)))


def mean_bias_error(y_true: ndarray, y_pred: ndarray) -> float:
    """Mean bias error regression loss.

    Args:
        y_true: Ground truth (correct) target values.
        y_pred: Estimated target values.
    """
    check_arrays(y_true, y_pred)
    return float(np.mean(y_pred - y_true))


def mean_squared_error(y_true: ndarray, y_pred: ndarray) -> float:
    """Mean squared error regression loss.

    Args:
        y_true: Ground truth (correct) target values.
        y_pred: Estimated target values.
    """
    check_arrays(y_true, y_pred)
    return float(np.mean((y_true - y_pred) ** 2))


def root_mean_squared_error(y_true: ndarray, y_pred: ndarray) -> float:
    """Root mean squared error regression loss.

    Args:
        y_true: Ground truth (correct) target values.
        y_pred: Estimated target values.
    """
    check_arrays(y_true, y_pred)
    return float(np.sqrt(np.mean((y_true - y_pred) ** 2)))


def mean_absolute_percentage_error(y_true: ndarray, y_pred: ndarray) -> float:
    """Mean absolute percentage error regression loss.

    Args:
        y_true: Ground truth (correct) target values.
        y_pred: Estimated target values.
    """
    check_arrays(y_true, y_pred)
    if np.any(y_true == 0):
        msg = "Found zero in true values. MAPE is undefined."
        raise ValueError(msg)
    return float(100.0 * np.mean(np.abs((y_true - y_pred) / y_true)))
